package parse;

import ast.*;
import java.io.IOException;
import java.util.ArrayList;

public class ScriptParser {
	public Scanner input;
	public Token la;
	
	public ScriptParser(Scanner input) {
		this.input = input;
	}
	
	public Token match(int token) {
		if(la.type == token) {
			Token tok = la;
			try {
				la = input.nextToken();
			} catch(IOException e) {
				throw new RuntimeException("erro na leitura da entrada: " + e.getMessage());
			}
			return tok;
		} else
			throw new RuntimeException("erro de sintaxe em " +
					la.line + ":" + la.col + ", esperado: "
					+ token + ", achado: " +
					la.type + "(" + la.val + ")");
	}
	
	public Exp parse() {
		try {
			la = input.nextToken();
			match(Token.LPE);
			return exp();
		} catch(IOException e) {
			throw new RuntimeException("erro na leitura da entrada: " + e.getMessage());
		}
	}
	
	public Bloco parseBloco() {
		try {
			la = input.nextToken();
			match(Token.LPE);
			return bloco();
		} catch(IOException e) {
			throw new RuntimeException("erro na leitura da entrada: " + e.getMessage());
		}
	}

	public Exp parseExp(Token la) {
		this.la = la;
		return exp();
	}

	
	// BLOCO -> {STAT}
	public Bloco bloco() {
		ArrayList<Stat> stats = new ArrayList<Stat>();
		int line = la.line; int col = la.col;
		// Conjunto FIRST(STAT)
		while(la.type == Token.DO || la.type == Token.IF ||
				la.type == Token.WHILE || la.type == '(' ||
				la.type == Token.ID || la.type == Token.LOCAL ||
				la.type == Token.FUNCTION) {
			stats.add(stat());
		}
		return new Bloco(stats, line, col);
	}

	// STAT -> do BLOCO end
	// STAT -> while EXP do BLOCO end
	// STAT -> if exp then BLOCO [else BLOCO] end
	// STAT -> local id [= EXP]
	// STAT -> LVAL = EXP
	public Stat stat() {
		switch(la.type) {
		case Token.DO: // STAT -> id = exp
			match(Token.DO);
			Bloco bloco = bloco();
			match(Token.END);
			return bloco;
		case Token.WHILE: { // STAT -> while EXP do BLOCO end
			Token tok = match(Token.WHILE);
			Exp cond = exp();
			match(Token.DO);
			Bloco b = bloco();
			match(Token.END);
			return new While(cond, b, tok.line, tok.col);
		}
		// STAT -> if exp then BLOCO [else BLOCO] end
		case Token.IF: {
			Token tok = match(Token.IF);
			Exp cond = exp();
			match(Token.THEN);
			Bloco bthen = bloco();
			Bloco belse = null;
			if(la.type == Token.ELSE) {
				match(Token.ELSE);
				belse = bloco();
			}
			match(Token.END);
			return new If(cond, bthen, belse, tok.line, tok.col);
		}
		case Token.LOCAL: { // STAT -> local id [= exp]
			Token tok = match(Token.LOCAL);
			Token tok_id = match(Token.ID);
			Id id = new Id(tok_id.val, tok_id.line, tok_id.col);
			Exp init = null;
			if(la.type == '=') {
				match('=');
				init = exp();
			}
			return new Local(id, init, tok.line, tok.col);
		}
		case Token.ID: {
			Id id = (Id)pexp();
			Token tok = match('=');
			Exp rval = exp();
			return new Atrib(id, rval, tok.line, tok.col);
		}
		default:
			throw new RuntimeException("comando invlido em " + la.line + ":" + la.col +
					", comeando com: " + la.type + "(" + la.val + ")");	
		}
	}

	public Exp exp() {
		return rexp();
	}
	
	// REXP   -> REXP ~= CEXP
	public Exp rexp() {
		Exp res = aexp();
		while(la.type == Token.NEQ) {
			Token tok = match(Token.NEQ);
			res = new Neq(res, aexp(), tok.line, tok.col);
		}
		return res;
	}
	
	// AEXP   -> AEXP + MEXP
	// AEXP   -> AEXP - MEXP
	// AEXP   -> MEXP
	public Exp aexp() {
		Exp res = mexp();
		while(la.type == '+' || la.type == '-') {
			if(la.type == '+') {
				Token tok = match('+');
				res = new Soma(res, mexp(), tok.line, tok.col);
			} else {
				Token tok = match('-');
				res = new Sub(res, mexp(), tok.line, tok.col);
			}
		}
		return res;
	}

	// MEXP   -> MEXP * SEXP
	// MEXP   -> MEXP / SEXP
	// MEXP   -> SEXP
	public Exp mexp() {
		Exp res = sexp();
		while(la.type == '*' || la.type == '/') {
			if(la.type == '*') {
				Token tok = match('*');
				res = new Mul(res, sexp(), tok.line, tok.col);
			} else {
				Token tok = match('/');
				res = new Div(res, sexp(), tok.line, tok.col);
			}
		}
		return res;
	}
	
	// SEXP -> num
	// SEXP -> PEXP
	public Exp sexp() {
		switch(la.type) {
		case Token.NUM:
			Token tok = match(Token.NUM);
			return new Num(tok.val, tok.line, tok.col);
		default:
			return pexp();
		}
	}

	// PEXP   -> id
	// PEXP   -> ( EXP )
	public Exp pexp() {
		switch(la.type) {
		case Token.ID:
			Token id = match(Token.ID);
			return new Id(id.val, id.line, id.col);
		case '(':
			match('(');
			Exp res = exp();
			match(')');
			return res;
		}
		return null;
	}

}